import array
from ms_deisotope.data_source.metadata import activation
import warnings

from abc import abstractmethod
from weakref import WeakValueDictionary
from itertools import chain

from collections import namedtuple, defaultdict

import numpy as np

from ms_deisotope.peak_set import IonMobilityDeconvolutedPeak, DeconvolutedPeakSet
from ms_deisotope.peak_dependency_network import IntervalTreeNode, Interval

from .base import RawDataArrays, ScanBunch

class IonMobilitySource(object):
    @abstractmethod
    def _frame_id(self, data):
        raise NotImplementedError()

    @abstractmethod
    def _frame_index(self, data):
        raise NotImplementedError()

    def _frame_time(self, data):
        scan = self.get_scan_by_index(
            self._frame_start_scan_index(data))
        return scan.scan_time

    def _frame_ms_level(self, data):
        scan = self.get_scan_by_index(
            self._frame_start_scan_index(data))
        return scan.ms_level

    @abstractmethod
    def _frame_start_scan_index(self, data):
        raise NotImplementedError()

    @abstractmethod
    def _frame_end_scan_index(self, data):
        raise NotImplementedError()

    def _frame_precursor_information(self, data):
        scan = self.get_scan_by_index(
            self._frame_start_scan_index(data))
        return scan.precursor_information

    def _frame_activation(self, data):
        scan = self.get_scan_by_index(
            self._frame_start_scan_index(data))
        return scan.activation

    def _frame_isolation_window(self, data):
        scan = self.get_scan_by_index(
            self._frame_start_scan_index(data))
        return scan.isolation_window

    def _frame_polarity(self, data):
        scan = self.get_scan_by_index(
            self._frame_start_scan_index(data))
        return scan.polarity

    def _frame_drift_times(self, data):
        scans = []
        for i in range(self._frame_start_scan_index(data), self._frame_end_scan_index(data)):
            scan = self.get_scan_by_index(i)
            scans.append(scan.drift_time)
        return scans

    def _frame_scans(self, data):
        scans = []
        for i in range(self._frame_start_scan_index(data), self._frame_end_scan_index(data)):
            scan = self.get_scan_by_index(i)
            scans.append(scan)
        return scans

    def _frame_annotations(self, data):
        return {}

    def _frame_arrays(self, data):
        scans = self._frame_scans(data)
        return RawDataArrays3D.stack(scans)

    def _frame_ion_mobilities(self, data):
        scans = self._frame_scans(data)
        return np.array([s.drift_time for s in scans])

    def _frame_acquisition_information(self, data):
        scan = self.get_scan_by_index(self._frame_start_scan_index(data))
        acq = scan.acquisition_information
        if acq is None:
            return acq
        acq = acq.copy()
        for event in acq:
            im = event._ion_mobility
            mt = im.ion_mobility_type()
            if mt is not None:
                im.remove_ion_mobility_type(mt)
        return acq


class IonMobilitySourceRandomAccessFrameSource(IonMobilitySource):
    @abstractmethod
    def get_frame_by_index(self, index):
        raise NotImplementedError()

    @abstractmethod
    def get_frame_by_time(self, time):
        raise NotImplementedError()

    def _validate_frame(self, data):
        return True

    def _make_frame(self, data):
        return IonMobilityFrame(data, self)

    def _cache_frame(self, frame):
        pass

    @abstractmethod
    def _default_frame_iterator(self, start_index=None):
        raise NotImplementedError()

    @abstractmethod
    def make_frame_iterator(self, iterator=None, grouped=False):
        raise NotImplementedError()

    @abstractmethod
    def start_from_frame(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True):
        raise NotImplementedError()

    def initialize_frame_cache(self):
        '''Initialize a cache which keeps track of which :class:`~.IonMobilityFrame`
        objects are still in memory using a :class:`weakref.WeakValueDictionary`.

        When a frame is requested, if the frame object is found in the cache, the
        existing object is returned rather than re-read from disk.
        '''
        self._frame_cache = WeakValueDictionary()

    @property
    def frame_cache(self):
        '''A :class:`weakref.WeakValueDictionary` mapping used to retrieve
        frames from memory if available before re-reading them from disk.
        '''
        return self._frame_cache

    @frame_cache.setter
    def frame_cache(self, value):
        self._frame_cache = value


class _upto_index(object):
    def __init__(self, values, last_value):
        self.values = values
        self.last_value = last_value
        self.index = np.searchsorted(self.values, last_value)

    def values_up_to(self, value):
        i = np.searchsorted(self.values, value)
        intermediate = self.values[self.index:i]
        self.index = i
        self.last_value = self.values[i - 1]
        return intermediate

    def __call__(self, value):
        return self.values_up_to(value)


class RawDataArrays3D(namedtuple("RawDataArrays3D", ['mz', 'intensity', 'ion_mobility'])):
    """Represent the m/z, intensity, and ion mobility arrays associated with a raw
    ion mobility frame of mass spectra.

    Thin wrapper around a ``namedtuple``, so this object supports
    the same interfaces as a tuple.

    Attributes
    ----------
    mz : :class:`np.ndarray`
        The m/z axis of a mass spectrum
    intensity : :class:`np.ndarray`
        The intensity measured at the corresponding m/z of a mass spectrum
    ion_mobility : :class:`np.ndarray`
        The ion mobility of each data point
    distinct_ion_mobility : :class:`np.ndarray`
        The distinct ion mobility values in the multidimensional space, independent
        of whether there are data points at that point
    ion_mobility_array_type : :class:`~.Term`
        The type of ion mobility array
    data_arrays : dict
        Any other data arrays
    """
    def __new__(cls, mz, intensity, ion_mobility, distinct_ion_mobility, ion_mobility_array_type=None, data_arrays=None):
        inst = super(RawDataArrays3D, cls).__new__(
            cls, mz, intensity, ion_mobility)
        inst.distinct_ion_mobility = distinct_ion_mobility
        inst.ion_mobility_array_type = ion_mobility_array_type
        inst.data_arrays = dict()
        if data_arrays:
            inst.data_arrays.update(data_arrays)
        return inst

    @property
    def drift_time(self):
        return self.ion_mobility

    def has_array(self, array_type):
        '''Check if this array set contains an array of the
        requested type.

        This method uses the semantic lookup mechanism to test
        "is-a" relationships so if a more abstract term is used,
        a wider range of terms may be matched.

        Parameters
        ----------
        array_type : str or :class:`~.Term`
            The array type name to test.

        Returns
        -------
        bool
        '''
        from ms_deisotope.data_source.metadata.scan_traits import binary_data_arrays
        try:
            term = binary_data_arrays[array_type]
        except KeyError:
            warnings.warn(
                "Array type %r could not be resolved, treating as a plain string" % (array_type, ))
            return array_type in self.binary_data_arrays
        if self.mz is not None and len(self.mz):
            k = binary_data_arrays['m/z array']
            if term.is_a(k):
                return k
        if self.intensity is not None and len(self.intensity):
            k = binary_data_arrays['intensity array']
            if term.is_a(k):
                return k
        if self.ion_mobility_array_type is not None:
            k = binary_data_arrays[self.ion_mobility_array_type]
            if term.is_a(k):
                return k
        for k in self.data_arrays:
            try:
                k = binary_data_arrays[k]
                if term.is_a(k):
                    return k
            except KeyError:
                if term == k:
                    return k
        return False

    def __copy__(self):
        inst = self.__class__(self.mz.copy(), self.intensity.copy(), self.ion_mobility.copy(),
                              self.distinct_ion_mobility.copy(), self.ion_mobility_array_type, {
            k: v.copy() for k, v in self.data_arrays.items()
        })
        return inst

    def copy(self):
        """Make a deep copy of this object.

        Returns
        -------
        :class:`RawDataArrays3D`
        """
        return self.__copy__()

    def __eq__(self, other):
        try:
            return np.allclose(
                self[0], other[0]) and np.allclose(
                    self[1], other[1]) and np.allclose(self[2], other[2])
        except ValueError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __mul__(self, i):
        return self.__class__(self.mz, self.intensity * i, self.ion_mobility,
                              self.distinct_ion_mobility, self.ion_mobility_array_type, self.data_arrays)

    def __div__(self, d):
        return self.__class__(self.mz, self.intensity / d, self.ion_mobility,
                              self.distinct_ion_mobility, self.ion_mobility_array_type, self.data_arrays)
    def __getitem__(self, i):
        if isinstance(i, int):
            return super(RawDataArrays3D, self).__getitem__(i)
        elif isinstance(i, (slice, list, tuple, np.ndarray)):
            return self._slice(i)
        else:
            return self.data_arrays[i]

    @classmethod
    def stack(cls, scans, ion_mobility_array_type=None):
        '''Combine multiple :class:`~.Scan` objects or (ion mobility, :class:`~.RawDataArrays`)
        pairs into a single :class:`~.RawDataArrays3D`

        Parameters
        ----------
        scans : list of :class:`~.Scan`
            The single ion mobility point arrays to combine
        ion_mobility_array_type : :class:`~.Term` or :class:`str`
            The type of ion mobility array
        '''
        mz = array.array('d')
        intensity = array.array('d')
        ion_mobility = array.array('d')
        data_arrays = defaultdict(lambda: array.array('d'))

        distinct_ion_mobility = array.array('d')
        for scan in scans:
            try:
                arrays = scan.arrays
                im = scan.drift_time
            except AttributeError:
                im = scan[0]
                arrays = scan[1]
            distinct_ion_mobility.append(im)
            if arrays.mz.size > 0:
                mz.extend(arrays.mz)
                intensity.extend(arrays.intensity)
                ion_mobility.extend([im] * arrays.mz.size)
                for key, val in arrays.data_arrays.items():
                    data_arrays[key].extend(val)
        mz = np.array(mz)
        ion_mobility = np.array(ion_mobility)
        mask = np.lexsort(np.stack((ion_mobility, mz,)))
        return cls(mz[mask], np.array(intensity)[mask], ion_mobility[mask],
                   np.sort(distinct_ion_mobility), ion_mobility_array_type,
                   {k: np.array(v)[mask] for k, v in data_arrays.items()})

    @classmethod
    def from_arrays(cls, mz_array, intensity_array, ion_mobility_array, ion_mobility_array_type=None, data_arrays=None):
        '''Build a new :class:`~.RawDataArrays3D` from parallel arrays.

        This will sort all arrays w.r.t. m/z and ion mobility.

        Parameters
        ----------
        mz_array : np.ndarray
            The m/z array
        intensity_array : np.ndarray
            The intensity array
        ion_mobility_array : np.ndarray
            The ion mobility array
        data_arrays : dict
            Any extra data arrays

        Returns
        -------
        :class:`~.RawDataArrays3D`
        '''
        mz_array = np.asarray(mz_array)
        ion_mobility_array = np.asarray(ion_mobility_array)
        distinct_ion_mobility = np.sort(np.unique(ion_mobility_array))
        mask = np.lexsort(np.stack((ion_mobility_array, mz_array)))
        if data_arrays:
            data_arrays = {
                k: np.asarray(v)[mask] for k, v in data_arrays.items()
            }
        return cls(
            mz_array[mask], np.asarray(intensity_array)[mask], ion_mobility_array[mask],
            distinct_ion_mobility, ion_mobility_array_type, data_arrays=data_arrays)


    @classmethod
    def empty(cls):
        return cls(np.array([]), np.array([]), np.array([]), np.array([]))

    def _slice(self, i, include_ion_mobility=True):
        mz = self.mz[i]
        intensity = self.intensity[i]
        data_arrays = {k: v[i] for k, v in self.data_arrays.items()}
        if not include_ion_mobility:
            return RawDataArrays(mz, intensity, data_arrays)
        return RawDataArrays3D(
            mz, intensity, self.ion_mobility[i], self.distinct_ion_mobility,
            self.ion_mobility_array_type, data_arrays)

    def unstack(self, include_empty=True):
        '''Convert this 3D array into a list of (ion mobility, :class:`~.RawDataArrays`) pairs

        Parameters
        ----------
        include_empty : bool
            Whether to include ion mobility values with no signal. Defaults to :const:`True`.

        Returns
        -------
        arrays : list[tuple[float, :class:`~.RawDataArrays`]]
            The split arrays arranged by drift time
        '''
        # Sort by m/z first, then by ion mobility
        mask = np.lexsort(np.stack((self.mz, self.ion_mobility)))
        sorted_dim = self.ion_mobility[mask]
        acc = []
        # re-sort all arrays along the mask, and then slice out in chunks.
        boundaries = np.where(np.diff(sorted_dim) != 0)[0]
        if len(boundaries) == 0:
            return [RawDataArrays(self.mz, self.intensity, self.data_arrays)]
        if boundaries[-1] < len(sorted_dim) - 1:
            boundaries = np.append(boundaries, [len(sorted_dim) - 1])
        last_boundary = 0
        index = _upto_index(self.distinct_ion_mobility, -1)
        for boundary in boundaries:
            drift_time = self.ion_mobility[mask[last_boundary]]
            for i in index(drift_time):
                acc.append((i, self._slice(slice(0, 0), False)))
            chunk = self._slice(mask[last_boundary:boundary + 1], False)
            acc.append((drift_time, chunk))
            last_boundary = boundary + 1
        return acc

    def grid(self):
        '''Convert this 3D array into an intensity grid with
        m/z along the rows and ion mobility along the columns.

        Returns
        -------
        mz_axis : np.ndarray
            The m/z value for each row
        im_axis : np.ndarray
            The ion mobility value for each column
        intensity : np.ndarray
            A 2D array of intensity values at each axis position
        '''
        mz_axis = np.unique(self.mz)
        im_axis = (self.distinct_ion_mobility)
        intensity = np.zeros((mz_axis.shape[0], im_axis.shape[0]))
        i1 = np.searchsorted(mz_axis, self.mz)
        i2 = np.searchsorted(im_axis, self.ion_mobility)
        intensity[i1, i2] += self.intensity
        return mz_axis, im_axis, intensity

    def find_mz(self, mz):
        """Find the nearest index to the query ``mz``

        Parameters
        ----------
        mz : float
            The m/z value to search for

        Returns
        -------
        int
            The index nearest to the query m/z
        """
        n = len(self.mz)
        lo = 0
        hi = n

        while hi != lo:
            mid = int((hi + lo) // 2)
            y = self.mz[mid]
            err = y - mz
            if abs(err) < 0.1:
                best_index = mid
                best_err = abs(err)
                i = mid
                while i >= 0:
                    y = self.mz[i]
                    err = y - mz
                    if err <= -0.1:
                        break
                    abs_err = abs(err)
                    if abs_err < best_err:
                        best_err = abs_err
                        best_index = i
                    i -= 1
                i = mid
                while i < n:
                    y = self.mz[i]
                    err = y - mz
                    if err >= 0.1:
                        break
                    abs_err = abs(err)
                    if abs_err < best_err:
                        best_err = abs_err
                        best_index = i
                    i += 1
                return best_index
            elif hi - lo == 1:
                return mid
            elif err > 0:
                hi = mid
            else:
                lo = mid
        return 0

    def between_mz(self, low, high):
        """Returns a slice of the arrays between ``low`` and ``high``
        m/z

        Parameters
        ----------
        low : float
            The lower bound m/z
        high : float
            The upper bound m/z

        Returns
        -------
        :class:`.RawDataArrays`
        """
        i = self.find_mz(low)
        mz_i = self.mz[i]
        while self.mz[i] == mz_i:
            i -= 1
        i += 1
        j = self.find_mz(high)
        mz_j = self.mz[j]
        while self.mz[j] == mz_j:
            j += 1
        if not (low <= self.mz[i] <= high):
            i += 1
        return self._slice(slice(i, j), True)


class FrameBase(object):

    @property
    def scan_time(self):
        return self.time

    @property
    def ion_mobilities(self):
        return self.source._frame_ion_mobilities(self._data)

    def scans(self):
        return self.source._frame_scans(self._data)

    def bind(self, source):
        self.source = source
        if self.precursor_information is not None:
            self.precursor_information.bind(source)
        return self

    def unbind(self):
        self.source = None
        if self.precursor_information:
            self.precursor_information.unbind()
        return self

    def __repr__(self):
        template = ("{self.__class__.__name__}({self.id}, index={self.index},"
                    " time={self.time}, ms_level={self.ms_level})")
        return template.format(self=self)

    def __eq__(self, other):
        if other is None:
            return False
        if self.id != other.id:
            return False
        elif self.ms_level != other.ms_level:
            return False
        elif self.time != other.time:
            return False
        elif self.deconvoluted_features != other.deconvoluted_features:
            return False
        elif self.features != other.features:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    __hash__ = None


class IonMobilityFrame(FrameBase):
    '''A :class:`IonMobilityFrame` represents an single time point acquisition of
    multiple mass spectra across multiple ion mobility drift time points. Because
    of its drift time by m/z structure, it does not have 1-D peak sets, but pseudo-
    :class:`~.LCMSFeatureMap`-like data structures which conserve the over-time
    property.
    '''
    def __init__(self, data, source):
        self._data = data
        self.source = source
        self.features = None
        self.deconvoluted_features = None

        self._unload()

    def _unload(self):
        self._id = None
        self._ms_level = None
        self._time = None
        self._polarity = None
        self._index = None
        self._start_scan_index = None
        self._end_scan_index = None

        self._arrays = None
        self._annotations = None

        self._isolation_window = None
        self._precursor_information = None
        self._activation = None
        self._acquisition_information = None

    @property
    def id(self):
        if self._id is None:
            self._id = self.source._frame_id(self._data)
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def index(self):
        if self._index is None:
            self._index = self.source._frame_index(self._data)
        return self._index

    @index.setter
    def index(self, value):
        self._index = value

    @property
    def time(self):
        if self._time is None:
            self._time = self.source._frame_time(self._data)
        return self._time

    @time.setter
    def time(self, value):
        self._time = value

    @property
    def ms_level(self):
        if self._ms_level is None:
            self._ms_level = self.source._frame_ms_level(self._data)
        return self._ms_level

    @ms_level.setter
    def ms_level(self, value):
        self._ms_level = value

    @property
    def polarity(self):
        if self._polarity is None:
            self._polarity = self.source._frame_polarity(self._data)
        return self._polarity

    @polarity.setter
    def polarity(self, value):
        self._polarity = value

    @property
    def start_scan_index(self):
        if self._start_scan_index is None:
            self._start_scan_index = self.source._frame_start_scan_index(self._data)
        return self._start_scan_index

    @start_scan_index.setter
    def start_scan_index(self, value):
        self._start_scan_index = value

    @property
    def end_scan_index(self):
        if self._end_scan_index is None:
            self._end_scan_index = self.source._frame_end_scan_index(self._data)
        return self._end_scan_index

    @end_scan_index.setter
    def end_scan_index(self, value):
        self._end_scan_index = value

    @property
    def precursor_information(self):
        if self._precursor_information is None:
            self._precursor_information = self.source._frame_precursor_information(self._data)
        return self._precursor_information

    @precursor_information.setter
    def precursor_information(self, value):
        self._precursor_information = value

    @property
    def activation(self):
        if self._activation is None:
            self._activation = self.source._frame_activation(self._data)
        return self._activation

    @activation.setter
    def activation(self, value):
        self._activation = value

    @property
    def isolation_window(self):
        if self._isolation_window is None:
            self._isolation_window = self.source._frame_isolation_window(self._data)
        return self._isolation_window

    @isolation_window.setter
    def isolation_window(self, value):
        self._isolation_window = value

    @property
    def acquisition_information(self):
        if self._acquisition_information is None:
            self._acquisition_information = self.source._frame_acquisition_information(self._data)
        return self._acquisition_information

    @acquisition_information.setter
    def acquisition_information(self, value):
        self._acquisition_information = value

    @property
    def arrays(self):
        if self._arrays is None:
            self._arrays = self.source._frame_arrays(self._data)
        return self._arrays

    @arrays.setter
    def arrays(self, value):
        self._arrays = value

    @property
    def annotations(self):
        if self._annotations is None:
            self._annotations = self.source._frame_annotations(self._data)
        return self._annotations

    @annotations.setter
    def annotations(self, value):
        self._annotations = dict(value)

    def get_scan_by_drift_time(self, drift_time):
        dt_axis = self.ion_mobilities
        lo = 0
        hi = n = len(dt_axis)
        best_match = None
        best_error = float('inf')

        # Handles infinities
        if drift_time > dt_axis[-1]:
            return self.source.get_scan_by_index(self.end_scan_index - 1)

        while hi != lo:
            mid = (hi + lo) // 2
            dt = dt_axis[mid]
            err = abs(dt - drift_time)
            if err < best_error:
                best_error = err
                best_match = mid
            if (dt == drift_time) or (hi - 1 == lo):
                i = mid - 1
                while i >= 0:
                    err = abs(dt_axis[i] - drift_time)
                    if err < best_error:
                        best_error = err
                        best_match = i
                        i -= 1
                    else:
                        break
                i = mid + 1
                while i < n:
                    err = abs(dt_axis[i] - drift_time)
                    if err < best_error:
                        best_error = err
                        best_match = i
                        i += 1
                    else:
                        break
                return self.source.get_scan_by_index(self.start_scan_index + best_match)
            elif dt > drift_time:
                hi = mid
            else:
                lo = mid

    def extract_features(self, error_tolerance=1.5e-5, max_gap_size=0.25, min_size=2, average=0, dx=0.001, **kwargs):
        from ms_deisotope.feature_map import feature_map
        scans = self.scans()
        lff = feature_map.LCMSFeatureForest(error_tolerance=error_tolerance)
        n = len(scans)
        for i, scan in enumerate(scans):
            if average:
                acc = []
                if i > 0:
                    for j in range(max((i - average, 0)), i):
                        acc.append(scans[j])
                    for j in range(i + 1, min(i + average + 1, n)):
                        acc.append(scans[j])
                scan = scan.average_with(acc, dx=dx)
            scan.pick_peaks(**kwargs)
            for peak in scan:
                lff.handle_peak(peak, scan.drift_time)
        lff.split_sparse(max_gap_size, min_size).smooth_overlaps(
            error_tolerance)
        self.features = lff
        return self

    def deconvolute_features(self, averagine=None, scorer=None, truncate_after=0.95,
                             minimum_intensity=5, min_size=2, max_gap_size=0.25, **kwargs):
        from ms_deisotope.feature_map import feature_processor
        if averagine is None:
            from ms_deisotope.averagine import peptide
            averagine = peptide
        if scorer is None:
            from ms_deisotope.scoring import MSDeconVFitter
            scorer = MSDeconVFitter(1)
        if self.features is None:
            raise ValueError(
                "IM-MS Features must be extracted before they can be charge state deconvoluted")
        decon = feature_processor.LCMSFeatureProcessor(
            self.features, averagine, scorer, minimum_size=min_size,
            maximum_time_gap=max_gap_size)
        self.deconvoluted_features = decon.deconvolute(
            minimum_intensity=minimum_intensity, truncate_after=truncate_after, **kwargs)
        return self

    def to_deconvoluted_peak_set(self, time_bound=None):
        if self.deconvoluted_features is None:
            raise ValueError("Must first deconvolute IMS features before converting to a time-binned peak set!")
        if time_bound is not None:
            cft = IntervalTreeNode.build(
                [Interval(f.start_time, f.end_time, [f]) for f in self.deconvoluted_features])
            features = chain.from_iterable(cft.overlaps(*time_bound))
        else:
            features = self.deconvoluted_features
        return features_to_peak_set(features)

    def pack(self, bind=False):
        packed = ProcessedIonMobilityFrame(
            self.id, self.precursor_information, self.ms_level,
            self.time, self.index, self.features, self.deconvoluted_features,
            self.start_scan_index, self.end_scan_index,
            self.polarity, self.activation, self.acquisition_information,
            self.isolation_window, self.annotations,
            self.source if bind else None)
        return packed


class ProcessedIonMobilityFrame(FrameBase):
    def __init__(self, id, precursor_information, ms_level, time, index,
                 features=None, deconvoluted_features=None,
                 start_scan_index=None, end_scan_index=None,
                 polarity=None, activation=None, acquisition_information=None,
                 isolation_window=None, annotations=None, source=None):
        if annotations is None:
            annotations = dict()
        self.source = source

        self.id = id
        self.precursor_information = precursor_information
        self.ms_level = ms_level
        self.time = time
        self.index = index
        self.features = features
        self.deconvoluted_features = deconvoluted_features

        self.start_scan_index = start_scan_index
        self.end_scan_index = end_scan_index
        self.polarity = polarity
        self.activation = activation
        self.acquisition_information = acquisition_information
        self.isolation_window = isolation_window
        self.annotations = annotations


def weighted_centroid(feature):
    total = 0
    normalizer = 0
    for node in feature:
        weight = node.total_intensity()
        total += node.time * weight
        normalizer += weight
    return total / normalizer


def merge_envelopes(envelopes):
    base = envelopes[0].clone()
    for env in envelopes[1:]:
        for i, p in enumerate(env):
            base[i].intensity += p.intensity

    return base


def feature_to_peak(feature):
    peak_cluster = feature.peaks
    peak_cluster = [pi for p in peak_cluster for pi in p]
    total_intensity = sum(p.intensity for p in peak_cluster)
    mz = sum(p.mz * p.intensity for p in peak_cluster) / total_intensity
    neutral_mass = sum(
        p.neutral_mass * p.intensity for p in peak_cluster) / total_intensity
    most_abundant_mass = sum(
        p.most_abundant_mass * p.intensity for p in peak_cluster) / total_intensity
    a_to_a2_ratio = sum(
        p.a_to_a2_ratio * p.intensity for p in peak_cluster) / total_intensity
    average_mass = sum(
        p.average_mass * p.intensity for p in peak_cluster) / total_intensity
    signal_to_noise = sum(p.signal_to_noise *
                          p.intensity for p in peak_cluster) / total_intensity
    fwhm = sum(p.full_width_at_half_max *
               p.intensity for p in peak_cluster) / total_intensity
    area = sum(p.area * p.intensity for p in peak_cluster) / total_intensity
    score = sum(p.score * p.intensity for p in peak_cluster) / total_intensity
    charge = peak_cluster[0].charge
    envelope = merge_envelopes([p.envelope for p in peak_cluster])
    return IonMobilityDeconvolutedPeak(
        neutral_mass=neutral_mass, intensity=total_intensity, charge=charge, signal_to_noise=signal_to_noise,
        full_width_at_half_max=fwhm, index=-1, a_to_a2_ratio=a_to_a2_ratio, most_abundant_mass=most_abundant_mass,
        average_mass=average_mass, score=score, envelope=envelope, mz=mz, fit=None, chosen_for_msms=False,
        area=area, drift_time=weighted_centroid(feature))


def features_to_peak_set(features):
    peak_set = DeconvolutedPeakSet(tuple(map(feature_to_peak, features)))
    peak_set.reindex()
    return peak_set


class Generic3DIonMobilityFrameSource(IonMobilitySourceRandomAccessFrameSource):
    def __init__(self, loader, **kwargs):
        self.loader = loader
        self.initialize_frame_cache()
        self._producer = self._wrap_iterator(self.loader)

    def _frame_acquisition_information(self, data):
        return self.loader._acquisition_information(data)

    def _frame_activation(self, data):
        return self.loader._activation(data)

    def _frame_id(self, data):
        return self.loader._scan_id(data)

    def _frame_index(self, data):
        return self.loader._scan_index(data)

    def _frame_time(self, data):
        return self.loader._scan_time(data)

    def _frame_precursor_information(self, data):
        return self.loader._precursor_information(data)

    def _frame_ms_level(self, data):
        return self.loader._ms_level(data)

    def _frame_polarity(self, data):
        return self.loader._polarity(data)

    def _frame_isolation_window(self, data):
        return self.loader._isolation_window(data)

    def _frame_start_scan_index(self, data):
        return self._frame_index(data)

    def _frame_end_scan_index(self, data):
        return self._frame_index(data) + 1

    def _frame_arrays(self, data):
        from ms_deisotope.data_source.metadata.scan_traits import binary_data_arrays

        mz, intensity, data_arrays = self.loader._scan_arrays(data)
        ion_mobility = None
        ion_mobility_type = None
        for key, value in data_arrays.items():
            try:
                acc = key.accession
                term = binary_data_arrays[acc]
            except (KeyError, AttributeError):
                continue
            if term.is_a('ion mobility array'):
                ion_mobility = value
                ion_mobility_type = key
                break
        if ion_mobility_type is None:
            raise ValueError(
                "Expected an ion mobility array type, but none were found.")
        d = data_arrays.copy()
        d.pop(ion_mobility_type)
        array3d = RawDataArrays3D.from_arrays(
            mz, intensity, ion_mobility, ion_mobility_type, d)
        return array3d

    def _frame_scans(self, data):
        from ms_deisotope.data_source.metadata.scan_traits import ion_mobility_attribute

        arrays = self._frame_arrays(data)
        scan_id_base = self._frame_id(data)
        i = 1
        scans = []
        for drift_time, arrays_i in arrays.unstack():
            subdat = data.copy()
            subdat['m/z array'] = arrays_i.mz
            subdat['intensity array'] = arrays_i.intensity
            subdat['id'] = scan_id_base + (' scan=%d' % i)
            i += 1
            scan = self.loader._make_scan(subdat)
            acq = scan.acquisition_information
            event = acq[0]
            tp = event.ion_mobility_type
            if tp:
                event._ion_mobility.remove_ion_mobility_type(tp)
            event._ion_mobility.add_ion_mobility(
                ion_mobility_attribute, drift_time)
            scans.append(scan)
        return scans

    def get_frame_by_index(self, index):
        scan = self.loader.get_scan_by_index(index)
        frame = self._make_frame(scan._data)
        self._cache_frame(frame)
        return frame

    def get_frame_by_id(self, id):
        scan = self.loader.get_scan_by_id(id)
        frame = self._make_frame(scan._data)
        self._cache_frame(frame)
        return frame

    def get_frame_by_time(self, time):
        scan = self.loader.get_scan_by_time(time)
        frame = self._make_frame(scan._data)
        self._cache_frame(frame)
        return frame

    def __len__(self):
        return len(self.loader)

    def __getitem__(self, i):
        result = self.loader[i]
        if isinstance(result, list):
            out = []
            for scan in result:
                frame = self._make_frame(scan._data)
                self._cache_frame(frame)
                out.append(frame)
            return out
        else:
            scan = result
            frame = self._make_frame(scan._data)
            self._cache_frame(frame)
            return frame

    def _wrap_iterator(self, iterator):
        for val in iterator:
            if isinstance(val, ScanBunch):
                precursor = self._make_frame(val.precursor._data)
                self._cache_frame(precursor)
                products = []
                for product in val.products:
                    product = self._make_frame(product._data)
                    self._cache_frame(product)
                    products.append(product)

                yield ScanBunch(precursor, products)
            else:
                frame = self._make_frame(val._data)
                self._cache_frame(frame)
                yield frame

    def _default_frame_iterator(self, start_index=None):
        self.loader.start_from_scan(index=start_index, grouped=False)
        return self._wrap_iterator(self.loader)

    def make_frame_iterator(self, iterator=None, grouped=False):
        self.loader.make_iterator(iterator, grouped=grouped)
        return self._wrap_iterator(self.loader)

    def start_from_frame(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True):
        self.loader.start_from_scan(
            scan_id=scan_id, rt=rt, index=index, require_ms1=require_ms1, grouped=grouped)
        self._producer = self._wrap_iterator(self.loader)
        return self

    def __next__(self):
        return next(self._producer)

    def next(self):
        return next(self._producer)

    def reset(self):
        self.loader.reset()
        self.initialize_frame_cache()
        return self


class FramedIonMobilityFrameSource(IonMobilitySourceRandomAccessFrameSource):
    def __init__(self, loader, **kwargs):
        self.loader = loader
        self.initialize_frame_cache()

    def _find_cycle_start(self, scan):
        dt = scan.drift_time
        if dt is None:
            return scan
        index = scan.index
        best_scan = scan
        while index >= 0:
            prev = self.loader[index]
            p_dt = prev.drift_time
            if p_dt is None:
                return best_scan
            if p_dt < dt:
                best_scan = scan
            else:
                return best_scan
            index -= 1
        return best_scan

    def _collect_cycle(self, scan):
        acc = []
        orig_scan = scan
        start_scan = self._find_cycle_start(scan)
        dt = start_scan.drift_time
        if dt is None:
            raise ValueError("Could not find a start scan with drift time from %r" % (scan, ))
        acc.append(start_scan)
        index = start_scan.index + 1
        while True:
            next_scan = self.loader.get_scan_by_index(index)
            next_dt = next_scan.drift_time
            if next_dt is None:
                break
            if next_dt < dt:
                break
            dt = next_dt
            acc.append(next_scan)
            index += 1
        return acc

    def _frame_id(self, frame):
        scans = self._frame_scans(frame)
        return ','.join([s.id for s in scans])

    def _frame_acquisition_information(self, data):
        return self.loader._acquisition_information(data)

    def _frame_activation(self, data):
        scans = self._frame_scans(data)
        return scans[0].activation

    def _frame_index(self, data):
        scans = self._frame_scans(data)
        return scans[0].index

    def _frame_time(self, data):
        scans = self._frame_scans(data)
        return scans[0].scan_time

    def _frame_ms_level(self, data):
        scans = self._frame_scans(data)
        return scans[0].ms_level

    def _frame_polarity(self, data):
        scans = self._frame_scans(data)
        return scans[0].polarity

    def _frame_isolation_window(self, data):
        scans = self._frame_scans(data)
        return scans[0].isolation_window

    def _frame_start_scan_index(self, data):
        scans = self._frame_scans(data)
        return scans[0].index

    def _frame_end_scan_index(self, data):
        scans = self._frame_scans(data)
        return scans[-1].index + 1

    def _frame_scans(self, data):
        try:
            return data['scans']
        except KeyError:
            pass
        source_id = data['_source_scan_id']
        scan = self.loader.get_scan_by_id(source_id)
        scans = self._collect_cycle(scan)
        data['scans'] = scans
        return data['scans']

    def get_scan_by_index(self, index):
        return self.loader.get_scan_by_index(index)



import logging
import array
from collections import defaultdict

import numpy as np

from ms_deisotope.data_source.common import ProcessedScan
from ms_deisotope import DeconvolutedPeakSet
from ms_peak_picker import PeakSet
from ms_deisotope.peak_dependency_network.intervals import Interval, IntervalTreeNode

from .lcms_feature import LCMSFeature
from .feature_fit import DeconvolutedLCMSFeature, IonMobilityDeconvolutedLCMSFeature


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

class _FeatureCollection(object):
    def __len__(self):
        return len(self.features)

    def __iter__(self):
        return iter(self.features)

    def __getitem__(self, i):
        if isinstance(i, (int, slice)):
            return self.features[i]
        else:
            return [self.features[j] for j in i]

    def __eq__(self, other):
        if other is None:
            return False
        return self.features == other.features

    def __ne__(self, other):
        return not self == other


class LCMSFeatureMap(_FeatureCollection):

    def __init__(self, features):
        self.features = sorted(features, key=lambda x: (x.mz, x.start_time))

    def search(self, mz, error_tolerance=2e-5):
        '''Search for a single feature within `error_tolerance` of `mz`.

        Parameters
        ----------
        mz : float
            The m/z to search for
        error_tolerance : float, optional
            The error tolerance to search with, defaults to 2e-5, 20 PPM

        Returns
        -------
        :class:`~.LCMSFeature`
            The a feature matching the query m/z, or :const:`None`.

        See Also
        --------
        find_all
        between
        '''
        i = binary_search(self.features, mz, error_tolerance)
        if i is None:
            return None
        match = self[i]
        return match

    def find_all(self, mz, error_tolerance=2e-5):
        '''Search for all features within `error_tolerance` of `mz`.

        Parameters
        ----------
        mz : float
            The m/z to search for
        error_tolerance : float, optional
            The error tolerance to search with, defaults to 2e-5, 20 PPM

        Returns
        -------
        list of :class:`~.LCMSFeature`
            A list of all features matching the query m/z, or the empty list.

        See Also
        --------
        between
        '''

        bounds = search_sweep(self.features, mz, error_tolerance)
        if bounds is not None:
            lo, hi = bounds
            return self[lo:hi]
        else:
            return []

    def spanning_time(self, time_point):
        return [feature for feature in self if feature.spans_in_time(time_point)]

    def index_range(self, lo, hi, error_tolerance=2e-5):
        return (
            binary_search_with_flag(
                self.features, lo, error_tolerance)[0][0],
            binary_search_with_flag(
                self.features, hi, error_tolerance)[0][0])

    def between(self, lo, hi, error_tolerance=2e-5):
        '''Search for all features between `lo` and `hi`, allowing `error_tolerance`
        around the edges.

        Parameters
        ----------
        lo : float
            The m/z to search for the lower bound
        hi : float
            The m/z to search for the upper bound
        error_tolerance : float, optional
            The error tolerance to search with, defaults to 2e-5, 20 PPM

        Returns
        -------
        list of :class:`~.LCMSFeature`
            The features between the query bounds with the permitted error, or an empty list.

        See Also
        --------
        find_all
        '''
        n = len(self)
        if n == 0:
            return []
        lo_ix = binary_search_with_flag(
            self.features, lo, error_tolerance)[0][0]
        if self[lo_ix].mz < lo:
            lo_ix += 1
        if lo_ix > len(self):
            lo_ix = len(self) - 1
        i = lo_ix
        while i < n:
            f = self[i]
            if f.mz > hi:
                break
            i += 1
        hi_ix = i
        return self[lo_ix:hi_ix]

    def __repr__(self):
        return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))

    def as_arrays(self):
        mz_array = array.array('d')
        intensity_array = array.array('d')
        ion_mobility_array = array.array('d')
        feature_id_array = array.array('L')
        for i, feature in enumerate(self):
            for node in feature:
                time = node.time
                for peak in node.members:
                    ion_mobility_array.append(time)
                    mz_array.append(peak.mz)
                    intensity_array.append(peak.intensity)
                    feature_id_array.append(i)
        mz_array = np.array(mz_array, copy=False)
        intensity_array = np.array(intensity_array, copy=False)
        ion_mobility_array = np.array(ion_mobility_array, copy=False)
        feature_id_array = np.array(feature_id_array, copy=False)
        mask = np.lexsort(np.stack((ion_mobility_array, mz_array)))
        return (mz_array[mask], intensity_array[mask], ion_mobility_array[mask], feature_id_array[mask])


try:
    from ms_deisotope._c.feature_map.feature_map import LCMSFeatureMap
except ImportError:
    pass


class LCMSFeatureForest(LCMSFeatureMap):
    """An an algorithm for aggregating features from peaks of similar m/z.

    Attributes
    ----------
    features : list of LCMSFeature
        A list of growing LCMSFeature objects, ordered by m/z
    count : int
        The number of peaks accumulated
    error_tolerance : float
        The mass error tolerance between peaks and possible features (in ppm)
    """
    def __init__(self, features=None, error_tolerance=1e-5):
        if features is None:
            features = []
        self.features = sorted(features, key=lambda x: x.mz)
        self.error_tolerance = error_tolerance
        self.count = 0

    def find_insertion_point(self, peak):
        '''Find the position in `self` where the peak's matching feature
        should be, and whether there was a feature that matched.

        Parameters
        ----------
        peak : PeakBase
            The peak to search for

        Returns
        -------
        index : int
            The position in `self` where the feature should be.
        matched : bool
            Whether there was a feature matched at that position
        '''
        index, matched = binary_search_with_flag(
            self.features, peak.mz, self.error_tolerance)
        return index, matched

    def find_minimizing_index(self, peak, indices):
        '''Amongst the set of `indices`, find the one that
        minimizes the distance to `peak`

        Parameters
        ----------
        peak : PeakBase
            The peak to match
        indicies : Iterable of int
            The feature inidices in `self` to choose between

        Returns
        -------
        int:
            The best index out of `indices`, with the smallest distance
            to `peak`
        '''
        best_index = None
        best_error = float('inf')
        for index_case in indices:
            feature = self[index_case]
            err = abs(feature.mz - peak.mz) / peak.mz
            if err < best_error:
                best_index = index_case
                best_error = err
        return best_index

    def handle_peak(self, peak, scan_time):
        '''Add `peak` at `scan_time` to the feature forest.

        If `peak` matches an existing feature, insert it into that feature
        with `scan_time`. If no feature is matched, create a new feature wit
        `peak` and add it to the feature forest.

        Parameters
        ----------
        peak : PeakBase
            The peak to add.
        scan_time : float
            The time the peak was observed.

        '''
        if len(self) == 0:
            index = [0]
            matched = False
        else:
            index, matched = self.find_insertion_point(peak)
        if matched:
            feature = self.features[self.find_minimizing_index(peak, index)]
            feature.insert(peak, scan_time)
        else:
            feature = LCMSFeature()
            feature.created_at = "forest"
            feature.insert(peak, scan_time)
            self.insert_feature(feature, index)
        self.count += 1

    def insert_feature(self, feature, index):
        '''Insert `feature` at `index` in :attr:`self.features`

        This method does minimal order-correctness checking.

        Parameters
        ----------
        feature : LCMSFeature
            The feature to be added.
        index : tuple of (int, bool)
            Where to insert the feature, and whether that feature had
            a match or not. By construction, index[1] should be :const`False`
            if this method is called.
        '''
        if index[0] != 0:
            self.features.insert(index[0] + 1, feature)
        else:
            if len(self) == 0:
                new_index = index[0]
            else:
                x = self.features[index[0]]
                if x.mz < feature.mz:
                    new_index = index[0] + 1
                else:
                    new_index = index[0]
            self.features.insert(new_index, feature)

    def split_sparse(self, delta_rt=1.0, min_size=2):
        '''Split features at gaps and eliminate features short features.

        This modifies `self` in-place.

        Parameters
        ----------
        delta_rt : float, optional
            The maximum gap size in the time dimension to tolerate. Defaults to 1.0 minute
        min_size : int, optional
            The minimum number of time points a feature must have in order to be retained.
            Defaults to 2.

        Returns
        -------
        self : LCMSFeatureForest
            This object, modified in-place.

        See Also
        --------
        LCMSFeature.split_sparse
        '''
        features = [fi for f in self.features for fi in f.split_sparse(delta_rt) if len(fi) >= min_size]
        self.features = features
        return self

    def smooth_overlaps(self, error_tolerance=None):
        '''Use a single pass of :func:`smooth_overlaps` to merge features which
        are nearby in the m/z dimension and overlap in the time dimension.

        This modifies `self` in-place.

        Parameters
        ----------
        error_tolerance : float, optional
            The mass error tolerance to use when finding similar features to merge.
            Defaults to :attr:`error_tolerance`

        Returns
        -------
        self : LCMSFeatureForest
            This object, modified in-place.

        See Also
        --------
        :func:`smooth_overlaps`
        :class:`LCMSFeatureOverlapSmoother`
        '''
        if error_tolerance is None:
            error_tolerance = self.error_tolerance
        self.features = smooth_overlaps(self.features, error_tolerance)
        return self

    def aggregate_peaks(self, scans, minimum_mz=160, minimum_intensity=500., maximum_mz=float('inf')):
        '''Aggregate peaks from `scans` into features in this collection.

        After all peaks are added, :meth:`smooth_overlaps` is used to merge
        nearby features.

        Parameters
        ----------
        scans : :class:`Iterable`[:class:`~.ScanBase`]
            The scan collection to aggregate from
        minimum_mz : float, optional
            The minimum peak m/z to be included. Defaults to 160 m/z
        minimum_intensity : float, optional
            The minimum peak intensity to be included. Defaults to 500.
        maximum_mz : float, optional
            The maximum peak m/z to be included. Defaults to infinity.

        Returns
        -------
        self : LCMSFeatureForest
            This object, modified in-place.
        '''
        for scan in scans:
            peak_set = scan.peak_set
            if peak_set is None:
                scan.pick_peaks()
                peak_set = scan.peak_set
            for peak in scan.peak_set:
                if peak.mz < minimum_mz or peak.mz > maximum_mz or peak.intensity < minimum_intensity:
                    continue
                self.handle_peak(peak, scan.scan_time)
        return self.smooth_overlaps()

    @classmethod
    def from_reader(cls, reader, error_tolerance=1e-5, minimum_mz=160, minimum_intensity=500., maximum_mz=float('inf'),
                    start_time=None, end_time=None):

        def generate():
            if start_time is not None:
                reader.start_from_scan(rt=start_time, grouped=False)
            else:
                reader.reset()
                reader.make_iterator(grouped=False)
            i = 0
            n = len(reader)
            for i, scan in enumerate(reader):
                if end_time is not None:
                    if scan.scan_time > end_time:
                        break
                logger.info("... Processed Scan %d/%d (%0.2f%%)" % (i, n, i * 100.0 / n))
                if scan.ms_level == 1:
                    yield scan

        self = cls(error_tolerance=error_tolerance)
        self.aggregate_peaks(
            generate(),
            minimum_mz=minimum_mz,
            minimum_intensity=minimum_intensity,
            maximum_mz=maximum_mz)
        return self


def smooth_overlaps_simple(feature_list, error_tolerance=1e-5):
    feature_list = sorted(feature_list, key=lambda x: x.mz)
    out = []
    last = feature_list[0]
    i = 1
    while i < len(feature_list):
        current = feature_list[i]
        mass_error = abs((last.mz - current.mz) / current.mz)
        if mass_error <= error_tolerance:
            if last.overlaps_in_time(current):
                last = last.merge(current)
                last.created_at = "smooth_overlaps"
            else:
                out.append(last)
                last = current
        else:
            out.append(last)
            last = current
        i += 1
    out.append(last)
    return out


def smooth_overlaps(feature_list, error_tolerance=1e-5, time_bridge=0):
    smoother = LCMSFeatureOverlapSmoother(feature_list, error_tolerance)
    return smoother.smooth()


def smooth_overlaps_neutral(feature_list, error_tolerance=1e-5, time_bridge=0):
    smoother = NeutralMassLCMSFeatureOverlapSmoother(feature_list, error_tolerance)
    return smoother.smooth()


def binary_search_with_flag(array, mz, error_tolerance=1e-5):
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.mz - mz) / mz
        if abs(err) <= error_tolerance:
            i = mid - 1
            # Begin Sweep forward
            while i > 0:
                x = array[i]
                err = (x.mz - mz) / mz
                if abs(err) <= error_tolerance:
                    i -= 1
                    continue
                else:
                    break
            low_end = i
            i = mid + 1

            # Begin Sweep backward
            while i < n:
                x = array[i]
                err = (x.mz - mz) / mz
                if abs(err) <= error_tolerance:
                    i += 1
                    continue
                else:
                    break
            high_end = i
            return list(range(low_end, high_end)), True
        elif (hi - lo) == 1:
            return [mid], False
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return [0], False


def binary_search(array, mz, error_tolerance=1e-5):
    """Binary search an ordered array of objects with :attr:`mz`
    using a PPM error tolerance of `error_tolerance`

    Parameters
    ----------
    array : list
        An list of objects, sorted over :attr:`mz` in increasing order
    mz : float
        The mz to search for
    error_tolerance : float, optional
        The PPM error tolerance to use when deciding whether a match has been found

    Returns
    -------
    int:
        The index in `array` of the best match
    """
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.mz - mz) / mz
        if abs(err) <= error_tolerance:
            best_index = mid
            best_error = abs(err)
            i = mid - 1
            while i >= 0:
                x = array[i]
                err = abs((x.mz - mz) / mz)
                if err < best_error:
                    best_error = err
                    best_index = i
                i -= 1

            i = mid + 1
            while i < n:
                x = array[i]
                err = abs((x.mz - mz) / mz)
                if err < best_error:
                    best_error = err
                    best_index = i
                i += 1
            return best_index
        elif (hi - lo) == 1:
            return None
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0


def binary_search_exact(array, mz):
    lo = 0
    hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.mz - mz)
        if err == 0:
            return mid
        elif (hi - lo) == 1:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0


def search_sweep(array, mz, error_tolerance=1e-5):
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.mz - mz) / mz
        if abs(err) <= error_tolerance:
            i = mid - 1
            while i >= 0:
                x = array[i]
                err = abs((x.mz - mz) / mz)
                if err > error_tolerance:
                    break
                i -= 1
            start = i + 1
            i = mid + 1
            while i < n:
                x = array[i]
                err = abs((x.mz - mz) / mz)
                if err > error_tolerance:
                    break
                i += 1
            end = i
            return (start, end)
        elif (hi - lo) == 1:
            return None
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0


def binary_search_with_flag_neutral(array, neutral_mass, error_tolerance=1e-5):
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.neutral_mass - neutral_mass) / neutral_mass
        if abs(err) <= error_tolerance:
            i = mid - 1
            # Begin Sweep forward
            while i > 0:
                x = array[i]
                err = (x.neutral_mass - neutral_mass) / neutral_mass
                if abs(err) <= error_tolerance:
                    i -= 1
                    continue
                else:
                    break
            low_end = i
            i = mid + 1

            # Begin Sweep backward
            while i < n:
                x = array[i]
                err = (x.neutral_mass - neutral_mass) / neutral_mass
                if abs(err) <= error_tolerance:
                    i += 1
                    continue
                else:
                    break
            high_end = i
            return list(range(low_end, high_end)), True
        elif (hi - lo) == 1:
            return [mid], False
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return [0], False


def binary_search_neutral(array, neutral_mass, error_tolerance=1e-5):
    """Binary search an ordered array of objects with :attr:`neutral_mass`
    using a PPM error tolerance of `error_toler

    Parameters
    ----------
    array : list
        An list of objects, sorted over :attr:`neutral_mass` in increasing order
    neutral_mass : float
        The neutral_mass to search for
    error_tolerance : float, optional
        The PPM error tolerance to use when deciding whether a match has been found

    Returns
    -------
    int:
        The index in `array` of the best match
    """
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.neutral_mass - neutral_mass) / neutral_mass
        if abs(err) <= error_tolerance:
            best_index = mid
            best_error = abs(err)
            i = mid - 1
            while i >= 0:
                x = array[i]
                err = abs((x.neutral_mass - neutral_mass) / neutral_mass)
                if err < best_error:
                    best_error = err
                    best_index = i
                i -= 1

            i = mid + 1
            while i < n:
                x = array[i]
                err = abs((x.neutral_mass - neutral_mass) / neutral_mass)
                if err < best_error:
                    best_error = err
                    best_index = i
                i += 1
            return best_index
        elif (hi - lo) == 1:
            return None
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0


def search_sweep_neutral(array, neutral_mass, error_tolerance=1e-5):
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.neutral_mass - neutral_mass) / neutral_mass
        if abs(err) <= error_tolerance:
            i = mid - 1
            while i >= 0:
                x = array[i]
                err = abs((x.neutral_mass - neutral_mass) / neutral_mass)
                if err > error_tolerance:
                    break
                i -= 1
            start = i + 1
            i = mid + 1
            while i < n:
                x = array[i]
                err = abs((x.neutral_mass - neutral_mass) / neutral_mass)
                if err > error_tolerance:
                    break
                i += 1
            end = i
            return (start, end)
        elif (hi - lo) == 1:
            return None
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0, 0


def binary_search_exact_neutral(array, neutral_mass):
    lo = 0
    hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.neutral_mass - neutral_mass)
        if err == 0:
            return mid
        elif (hi - lo) == 1:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0


class DeconvolutedLCMSFeatureMap(_FeatureCollection):

    def __init__(self, features):
        self.features = sorted(features, key=lambda x: (round(x.neutral_mass, 6), x.start_time))
        self._by_mz = sorted(features, key=lambda x: (x.mz, x.start_time))

    def spanning_time(self, time_point):
        return [feature for feature in self if feature.spans_in_time(time_point)]

    def search(self, mass, error_tolerance=2e-5, use_mz=False):
        '''Search for a single feature within `error_tolerance` of `mass`.

        By default, this searches in the neutral mass domain, to search in
        the m/z domain, pass `use_mz=True`.

        Parameters
        ----------
        mass : float
            The mass or m/z to search for
        error_tolerance : float, optional
            The error tolerance to search with, defaults to 2e-5, 20 PPM
        use_mz : bool
            Whether `mass` is interpreted as neutral or mass to charge ratio

        Returns
        -------
        :class:`~.LCMSFeature`
            The a feature matching the query mass, or :const:`None`.

        See Also
        --------
        find_all
        between
        '''
        if use_mz:
            i = binary_search(self._by_mz, mass, error_tolerance)
            if i is None:
                return None
            return self._by_mz[i]
        i = binary_search_neutral(self.features, mass, error_tolerance)
        if i is None:
            return None
        match = self[i]
        return match

    def find_all(self, mass, error_tolerance=2e-5, use_mz=False):
        '''Search for all features within `error_tolerance` of `mass`.

        By default, this searches in the neutral mass domain, to search in
        the m/z domain, pass `use_mz=True`.

        Parameters
        ----------
        mass : float
            The mass or m/z to search for
        error_tolerance : float, optional
            The error tolerance to search with, defaults to 2e-5, 20 PPM
        use_mz : bool
            Whether `mass` is interpreted as neutral or mass to charge ratio

        Returns
        -------
        list of :class:`~.LCMSFeature`
            The features matching the query mass with the permitted error, or an empty list.

        See Also
        --------
        between
        search
        '''
        if use_mz:
            bounds = search_sweep(self._by_mz, mass, error_tolerance)
            collection = self._by_mz
        else:
            bounds = search_sweep_neutral(self.features, mass, error_tolerance)
            collection = self
        if bounds is not None:
            lo, hi = bounds
            return collection[lo:hi]
        else:
            return []

    def index_range(self, lo, hi, error_tolerance=2e-5):
        return (
            binary_search_with_flag_neutral(
                self.features, lo, error_tolerance)[0][0],
            binary_search_with_flag_neutral(
                self.features, hi, error_tolerance)[0][0])

    def between(self, lo, hi, error_tolerance=2e-5, use_mz=False):
        '''Search for all features between `lo` and `hi`, allowing `error_tolerance`
        around the edges.

        By default, this searches in the neutral mass domain, to search in
        the m/z domain, pass `use_mz=True`.

        Parameters
        ----------
        lo : float
            The mass or m/z to search for the lower bound
        hi : float
            The mass or m/z to search for the upper bound
        error_tolerance : float, optional
            The error tolerance to search with, defaults to 2e-5, 20 PPM
        use_mz : bool
            Whether `lo` and `hi` are interpreted as neutral or mass to charge ratio

        Returns
        -------
        list of :class:`~.LCMSFeature`
            The features between the query bounds with the permitted error, or an empty list.

        See Also
        --------
        find_all
        '''
        n = len(self)
        if n == 0:
            return []
        if use_mz:
            lo_ix = binary_search_with_flag(
                self._by_mz, lo, error_tolerance)[0][0]
            if self._by_mz[lo_ix].mz < lo:
                lo_ix += 1
            if lo_ix > len(self):
                lo_ix = len(self) - 1
            i = lo_ix
            while i < n:
                f = self._by_mz[i]
                if f.mz > hi:
                    break
                i += 1
            hi_ix = i
            return self._by_mz[lo_ix:hi_ix]
        else:
            lo_ix = binary_search_with_flag_neutral(
                self.features, lo, error_tolerance)[0][0]
            if self[lo_ix].neutral_mass < lo:
                lo_ix += 1
            if lo_ix > len(self):
                lo_ix = len(self) - 1
            i = lo_ix
            while i < n:
                f = self[i]
                if f.neutral_mass > hi:
                    break
                i += 1
            hi_ix = i
            return self[lo_ix:hi_ix]

    def __repr__(self):
        return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))

    def as_arrays(self):
        mz_array = array.array('d')
        intensity_array = array.array('d')
        charge_array = array.array('i')
        score_array = array.array('d')
        ion_mobility_array = array.array('d')
        feature_id_array = array.array('L')
        envelopes = []

        point_count = 0
        for i, feature in enumerate(self):
            for node in feature:
                time = node.time
                for peak in node.members:
                    ion_mobility_array.append(time)
                    mz_array.append(peak.mz)
                    intensity_array.append(peak.intensity)
                    score_array.append(peak.score)
                    charge_array.append(peak.charge)
                    envelopes.append(peak.envelope)
                    point_count += (len(peak.envelope) + 1) * 2
                    feature_id_array.append(i)

        mz_array = np.array(mz_array, copy=False)
        intensity_array = np.array(intensity_array, copy=False)
        charge_array = np.array(charge_array, copy=False)
        score_array = np.array(score_array, copy=False)
        ion_mobility_array = np.array(ion_mobility_array, copy=False)
        feature_id_array = np.array(feature_id_array, copy=False)
        mask = np.lexsort(np.stack((ion_mobility_array, mz_array)))
        envelope_array = np.zeros(point_count, dtype=np.float32)

        k = 0
        for j in mask:
            for point in envelopes[j]:
                envelope_array[k] = point.mz
                envelope_array[k+1] = point.intensity
                k += 2
            k += 2
        return (mz_array[mask], intensity_array[mask], charge_array[mask],
                score_array[mask], ion_mobility_array[mask], envelope_array,
                feature_id_array[mask])


def convert_map_to_scan_peak_list(feature_map, peak_loader, time_precision=4, deconvoluted=True):
    metadata_map = {}
    scan_accumulator = defaultdict(list)
    for scan_id, metadata in peak_loader.extended_index.ms1_ids.items():
        metadata_map[round(metadata["scan_time"], time_precision)] = metadata
    for feature in feature_map:
        for node in feature:
            scan_accumulator[round(node.time, time_precision)].extend(node.members)

    packed = []
    for key, peaks in sorted(scan_accumulator.items(), key=lambda x: x[0]):
        template = peak_loader.get_scan_by_time(key)
        if deconvoluted:
            peak_set = PeakSet([])
            deconvoluted_peak_set = DeconvolutedPeakSet(peaks)
        else:
            peak_set = PeakSet(peaks)
            deconvoluted_peak_set = DeconvolutedPeakSet([])
        peak_set.reindex()
        deconvoluted_peak_set.reindex()
        scan = ProcessedScan(
            template.id, template.title, None, template.ms_level, template.scan_time,
            template.index, peak_set, deconvoluted_peak_set, template.polarity,
            None)
        packed.append(scan)
    return packed


class _FeatureIndex(object):
    def __iter__(self):
        return iter(self.features)

    def __getitem__(self, i):
        return self.features[i]

    def __len__(self):
        return len(self.features)

    def __nonzero__(self):
        return bool(self.features)

    def __bool__(self):
        return bool(self.features)

    def __repr__(self):
        return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))


class MZIndex(_FeatureIndex):
    def __init__(self, features):
        self.features = sorted(features, key=lambda x: x.mz)

    def find_all(self, mz, error_tolerance=2e-5):
        bounds = search_sweep(self.features, mz, error_tolerance)
        if bounds is not None:
            lo, hi = bounds
            return self[lo:hi]
        else:
            return []


class NeutralMassIndex(_FeatureIndex):
    def __init__(self, features):
        self.features = sorted(features, key=lambda x: x.neutral_mass)

    def find_all(self, mass, error_tolerance=2e-5):
        bounds = search_sweep_neutral(self.features, mass, error_tolerance)
        if bounds is not None:
            lo, hi = bounds
            return self[lo:hi]
        else:
            return []


try:
    from ms_deisotope._c.feature_map.feature_map import binary_search_with_flag
except ImportError:
    pass


class FeatureRetentionTimeInterval(Interval):
    def __init__(self, chromatogram):
        super(FeatureRetentionTimeInterval, self).__init__(
            chromatogram.start_time, chromatogram.end_time, [chromatogram])
        self.neutral_mass = chromatogram.neutral_mass
        self.start_time = self.start
        self.end_time = self.end
        self.data['neutral_mass'] = self.neutral_mass


def build_rt_interval_tree(chromatogram_list, interval_tree_type=IntervalTreeNode):
    intervals = list(map(FeatureRetentionTimeInterval, chromatogram_list))
    interval_tree = interval_tree_type.build(intervals)
    return interval_tree


class LCMSFeatureMerger(_FeatureCollection):
    def __init__(self, features=None, error_tolerance=1e-5):
        if features is None:
            features = []
        self.features = sorted(
            features, key=self._mass_coordinate)
        self.error_tolerance = error_tolerance
        self.count = 0

    def _mass_coordinate(self, x):
        return x.mz

    # Mass Coordinate Explicit
    def find_candidates(self, new_feature):
        index, matched = binary_search_with_flag(
            self.features, new_feature.mz, self.error_tolerance)
        return index, matched

    def merge_overlaps(self, new_feature, feature_range):
        has_merged = False
        query_mass = self._mass_coordinate(new_feature)

        for chroma in feature_range:
            cond = chroma.overlaps_in_time(new_feature)
            cond = (cond and abs(
                    (self._mass_coordinate(chroma) - query_mass) / query_mass) < self.error_tolerance)
            if cond:
                chroma.merge(new_feature)
                has_merged = True
                break
        return has_merged

    # Mass Coordinate Explicit
    def find_insertion_point(self, new_feature):
        return binary_search_exact(
            self.features, new_feature.mz)

    def handle_new_feature(self, new_feature):
        if len(self) == 0:
            index = [0]
            matched = False
        else:
            index, matched = self.find_candidates(new_feature)
        if matched:

            chroma = self[index]
            has_merged = self.merge_overlaps(new_feature, chroma)
            if not has_merged:
                insertion_point = self.find_insertion_point(new_feature)
                self.insert_feature(new_feature, insertion_point)
        else:
            self.insert_feature(new_feature, index[0])
        self.count += 1

    def insert_feature(self, feature, index):
        if index != 0:
            self.features.insert(index + 1, feature)
        else:
            if len(self) == 0:
                new_index = index
            else:
                x = self.features[index]
                if self._mass_coordinate(x) < self._mass_coordinate(feature):
                    new_index = index + 1
                else:
                    new_index = index
            self.features.insert(new_index, feature)

    def aggregate_features(self, features):
        unmatched = sorted(
            features, key=lambda x: x.total_signal, reverse=True)
        for chroma in unmatched:
            self.handle_new_feature(chroma)


def flatten_tree(tree):
    '''Perform a FIFO infix right-to-left traversal of the tree
    to flatten an :class:`~.IntervalTreeNode` hierarchy into a
    list.

    This still iterates over interval tree nodes, unlike the
    IntervalTreeNode class's own `__iter__` method, which flattens
    out the tree to iterate over contained intervals.

    Parameters
    ----------
    tree: :class:`~.IntervalTreeNode`

    Returns
    -------
    list
    '''
    output_queue = []
    input_queue = [tree]
    while input_queue:
        next_node = input_queue.pop()
        output_queue.append(next_node)

        next_right = next_node.right
        if next_right is not None:
            input_queue.append(next_right)

        next_left = next_node.left
        if next_left is not None:
            input_queue.append(next_left)
    return output_queue[::-1]


def layered_traversal(nodes):
    return sorted(nodes, key=lambda x: (x.level, x.center), reverse=True)


class LCMSFeatureOverlapSmoother(_FeatureCollection):
    '''Merge features which overlap in time but which are within :attr:`error_tolerance`
    mass units of another, recursively merging upwards from smallest to largest intervals.

    Attributes
    ----------
    '''
    def __init__(self, features, error_tolerance=1e-5, time_bridge=0):
        self.retention_interval_tree = build_rt_interval_tree(features)
        self.error_tolerance = error_tolerance
        self.time_bridge = time_bridge
        self.solution_map = {None: []}
        self.features = self.smooth()

    def _merge_features(self, features):
        merger = LCMSFeatureMerger(error_tolerance=self.error_tolerance)
        merger.aggregate_features(features)
        return list(merger)

    def aggregate_interval(self, tree):
        features = [interval[0] for interval in tree.contained]
        features.extend(self.solution_map[tree.left])
        features.extend(self.solution_map[tree.right])
        merged = self.solution_map[tree] = self._merge_features(features)
        return merged

    def smooth(self):
        nodes = layered_traversal(flatten_tree(self.retention_interval_tree))
        for node in nodes:
            self.aggregate_interval(node)
        final = self.solution_map[self.retention_interval_tree]
        result = self._merge_features(final)
        return result


class NeutralMassLCMSFeatureMerger(LCMSFeatureMerger):
    def _mass_coordinate(self, x):
        return x.neutral_mass

    # Mass Coordinate Explicit
    def find_candidates(self, new_feature):
        index, matched = binary_search_with_flag_neutral(
            self.features, new_feature.neutral_mass, self.error_tolerance)
        return index, matched

    def merge_overlaps(self, new_feature, feature_range):
        has_merged = False
        query_mass = self._mass_coordinate(new_feature)
        for chroma in feature_range:
            cond = chroma.overlaps_in_time(new_feature)
            cond = (cond and chroma.charge == new_feature.charge and abs(
                    (self._mass_coordinate(chroma) - query_mass) / query_mass) < self.error_tolerance)
            if cond:
                chroma.merge(new_feature)
                has_merged = True
                break
        return has_merged

    # Mass Coordinate Explicit
    def find_insertion_point(self, new_feature):
        return binary_search_exact_neutral(
            self.features, new_feature.neutral_mass)


class NeutralMassLCMSFeatureOverlapSmoother(LCMSFeatureOverlapSmoother):
    def _merge_features(self, features):
        merger = NeutralMassLCMSFeatureMerger(error_tolerance=self.error_tolerance)
        merger.aggregate_features(features)
        return list(merger)


class DeconvolutedLCMSFeatureForest(DeconvolutedLCMSFeatureMap):
    """An algorithm for aggregating features from peaks of close mass.

    Attributes
    ----------
    features : list of LCMSFeature
        A list of growing LCMSFeature objects, ordered by neutral mass
    count : int
        The number of peaks accumulated
    error_tolerance : float
        The mass error tolerance between peaks and possible features (in ppm)
    """

    def __init__(self, features=None, error_tolerance=1e-5):
        if features is None:
            features = []
        self.features = sorted(features, key=lambda x: x.neutral_mass)
        self._by_mz = sorted(features, key=lambda x: x.mz)
        self.error_tolerance = error_tolerance
        self.count = 0


    def find_insertion_point(self, peak):
        '''Find the position in `self` where the peak's matching feature
        should be, and whether there was a feature that matched.

        Parameters
        ----------
        peak : DeconvolutedPeak
            The peak to search for

        Returns
        -------
        index : int
            The position in `self` where the feature should be.
        matched : bool
            Whether there was a feature matched at that position
        '''
        index, matched = binary_search_with_flag_neutral(
            self.features, peak.neutral_mass, self.error_tolerance)
        return index, matched

    def find_minimizing_index(self, peak, indices):
        '''Amongst the set of `indices`, find the one that
        minimizes the distance to `peak`

        Parameters
        ----------
        peak : PeakBase
            The peak to match
        indicies : Iterable of int
            The feature inidices in `self` to choose between

        Returns
        -------
        int:
            The best index out of `indices`, with the smallest distance
            to `peak`
        '''
        best_index = None
        best_error = float('inf')
        for index_case in indices:
            feature = self[index_case]
            if feature.charge != peak.charge:
                continue
            err = abs(feature.neutral_mass - peak.neutral_mass) / \
                peak.neutral_mass
            if err < best_error:
                best_index = index_case
                best_error = err
        return best_index

    def handle_peak(self, peak, scan_time):
        '''Add `peak` at `scan_time` to the feature forest.

        If `peak` matches an existing feature, insert it into that feature
        with `scan_time`. If no feature is matched, create a new feature wit
        `peak` and add it to the feature forest.

        Parameters
        ----------
        peak : DeconvolutedPeak
            The peak to add.
        scan_time : float
            The time the peak was observed.

        '''
        if len(self) == 0:
            index = [0]
            matched = False
        else:
            index, matched = self.find_insertion_point(peak)
        if matched:
            minimized_feature_index = self.find_minimizing_index(peak, index)
            if minimized_feature_index is not None:
                feature = self.features[minimized_feature_index]
                feature.insert(peak, scan_time)
            else:
                feature = DeconvolutedLCMSFeature(charge=peak.charge)
                feature.created_at = "forest"
                feature.insert(peak, scan_time)
                self.insert_feature(feature, index)
        else:
            feature = DeconvolutedLCMSFeature(charge=peak.charge)
            feature.created_at = "forest"
            feature.insert(peak, scan_time)
            self.insert_feature(feature, index)
        self.count += 1

    def insert_feature(self, feature, index):
        '''Insert `feature` at `index` in :attr:`self.features`

        This method does minimal order-correctness checking.

        Parameters
        ----------
        feature : DeconvolutedLCMSFeature
            The feature to be added.
        index : tuple of (int, bool)
            Where to insert the feature, and whether that feature had
            a match or not. By construction, index[1] should be :const`False`
            if this method is called.
        '''
        if index[0] != 0:
            self.features.insert(index[0] + 1, feature)
        else:
            if len(self) == 0:
                new_index = index[0]
            else:
                x = self.features[index[0]]
                if x.neutral_mass < feature.neutral_mass:
                    new_index = index[0] + 1
                else:
                    new_index = index[0]
            self.features.insert(new_index, feature)

    def split_sparse(self, delta_rt=1.0, min_size=2):
        '''Split features at gaps and eliminate features short features.

        This modifies `self` in-place.

        Parameters
        ----------
        delta_rt : float, optional
            The maximum gap size in the time dimension to tolerate. Defaults to 1.0 minute
        min_size : int, optional
            The minimum number of time points a feature must have in order to be retained.
            Defaults to 2.

        Returns
        -------
        self : DeconvolutedLCMSFeatureForest
            This object, modified in-place.

        See Also
        --------
        DeconvolutedLCMSFeature.split_sparse
        '''
        features = [fi for f in self.features for fi in f.split_sparse(
            delta_rt) if len(fi) >= min_size]
        self.features = features
        return self

    def smooth_overlaps(self, error_tolerance=None):
        '''Use a single pass of :func:`smooth_overlaps_neutral` to merge features which
        are nearby in the neutral mass dimension and overlap in the time dimension.

        This modifies `self` in-place.

        Parameters
        ----------
        error_tolerance : float, optional
            The mass error tolerance to use when finding similar features to merge.
            Defaults to :attr:`error_tolerance`

        Returns
        -------
        self : DeconvolutedLCMSFeatureForest
            This object, modified in-place.

        See Also
        --------
        :func:`smooth_overlaps_neutral`
        :class:`NeutralMassLCMSFeatureOverlapSmoother`
        '''
        if error_tolerance is None:
            error_tolerance = self.error_tolerance
        self.features = smooth_overlaps_neutral(self.features, error_tolerance)
        self._by_mz = sorted(self.features, key=lambda x: x.mz)
        return self

    def aggregate_peaks(self, scans, minimum_mass=160, minimum_intensity=10., maximum_mass=float('inf')):
        '''Aggregate peaks from `scans` into features in this collection.

        After all peaks are added, :meth:`smooth_overlaps` is used to merge
        nearby features.

        Parameters
        ----------
        scans : :class:`Iterable`[:class:`~.ScanBase`]
            The scan collection to aggregate from
        minimum_mass : float, optional
            The minimum peak neutral mass to be included. Defaults to 160 Da
        minimum_intensity : float, optional
            The minimum peak intensity to be included. Defaults to 10.
        maximum_mass : float, optional
            The maximum peak neutral mass to be included. Defaults to infinity.

        Returns
        -------
        self : DeconvolutedLCMSFeatureForest
            This object, modified in-place.
        '''
        for scan in scans:
            peak_set = scan.deconvoluted_peak_set
            if peak_set is None:
                scan.pick_peaks()
                peak_set = scan.deconvoluted_peak_set
            for peak in peak_set:
                if peak.neutral_mass < minimum_mass or peak.neutral_mass > maximum_mass or peak.intensity < minimum_intensity:
                    continue
                self.handle_peak(peak, scan.scan_time)
        self.smooth_overlaps()
        return self

    @classmethod
    def from_reader(cls, reader, error_tolerance=1e-5, minimum_mass=160, minimum_intensity=10., maximum_mass=float('inf'),
                    start_time=None, end_time=None, ms_level=1):

        def generate():
            if start_time is not None:
                reader.start_from_scan(rt=start_time, grouped=False)
            else:
                reader.reset()
                reader.make_iterator(grouped=False)
            i = 0
            n = len(reader)
            for i, scan in enumerate(reader):
                if end_time is not None:
                    if scan.scan_time > end_time:
                        break
                    logger.info("... Processed Scan %d/%d (%0.2f%%)" %
                                (i, n, i * 100.0 / n))
                if scan.ms_level == ms_level:
                    yield scan

        self = cls(error_tolerance=error_tolerance)
        self.aggregate_peaks(
            generate(),
            minimum_mass=minimum_mass,
            minimum_intensity=minimum_intensity,
            maximum_mass=maximum_mass)
        return self


class IonMobilityNeutralMassLCMSFeatureMerger(NeutralMassLCMSFeatureMerger):
    def __init__(self, features=None, error_tolerance=1e-5, drift_error_tolerance=0.1):
        self.drift_error_tolerance = drift_error_tolerance
        super(IonMobilityNeutralMassLCMSFeatureMerger, self).__init__(features, error_tolerance)

    def merge_overlaps(self, new_feature, feature_range):
        has_merged = False
        query_mass = self._mass_coordinate(new_feature)
        for chroma in feature_range:
            cond = chroma.overlaps_in_time(new_feature)
            cond = (cond and chroma.charge == new_feature.charge and abs(
                    (self._mass_coordinate(chroma) - query_mass) / query_mass) < self.error_tolerance)
            cond = (cond and abs(chroma.drift_time - new_feature.drift_time) < self.drift_error_tolerance)
            if cond:
                chroma.merge(new_feature)
                has_merged = True
                break
        return has_merged


class IonMobilityNeutralMassLCMSFeatureOverlapSmoother(NeutralMassLCMSFeatureOverlapSmoother):
    def __init__(self, features, error_tolerance=1e-5, time_bridge=0, drift_error_tolerance=0.1):
        self.drift_error_tolerance = drift_error_tolerance
        super(IonMobilityNeutralMassLCMSFeatureOverlapSmoother, self).__init__(features, error_tolerance, time_bridge)

    def _merge_features(self, features):
        merger = IonMobilityNeutralMassLCMSFeatureMerger(
            error_tolerance=self.error_tolerance, drift_error_tolerance=self.drift_error_tolerance)
        merger.aggregate_features(features)
        return list(merger)


def smooth_overlaps_neutral_ims(feature_list, error_tolerance=1e-5, time_bridge=0, drift_error_tolerance=0.1):
    smoother = IonMobilityNeutralMassLCMSFeatureOverlapSmoother(
        feature_list, error_tolerance, drift_error_tolerance=drift_error_tolerance)
    return smoother.smooth()


class IonMobilityDeconvolutedLCMSFeatureForest(DeconvolutedLCMSFeatureForest):
    def __init__(self, features=None, error_tolerance=1e-5, drift_error_tolerance=0.1):
        self.drift_error_tolerance = drift_error_tolerance
        super(IonMobilityDeconvolutedLCMSFeatureForest, self).__init__(features, error_tolerance)

    def smooth_overlaps(self, error_tolerance=None):
        if error_tolerance is None:
            error_tolerance = self.error_tolerance
        self.features = smooth_overlaps_neutral_ims(
            self.features, error_tolerance, drift_error_tolerance=self.drift_error_tolerance)
        return self

    def find_minimizing_index(self, peak, indices):
        best_index = None
        best_error = float('inf')
        for index_case in indices:
            feature = self[index_case]
            if feature.charge != peak.charge:
                continue
            if abs(feature.drift_time - peak.drift_time) >= self.drift_error_tolerance:
                continue
            err = abs(feature.neutral_mass - peak.neutral_mass) / peak.neutral_mass
            if err < best_error:
                best_index = index_case
                best_error = err
        return best_index

    def handle_peak(self, peak, scan_time):
        if len(self) == 0:
            index = [0]
            matched = False
        else:
            index, matched = self.find_insertion_point(peak)
        if matched:
            minimized_feature_index = self.find_minimizing_index(peak, index)
            if minimized_feature_index is not None:
                feature = self.features[minimized_feature_index]
                feature.insert(peak, scan_time)
            else:
                feature = IonMobilityDeconvolutedLCMSFeature(charge=peak.charge)
                feature.created_at = "forest"
                feature.insert(peak, scan_time)
                self.insert_feature(feature, index)
        else:
            feature = IonMobilityDeconvolutedLCMSFeature(charge=peak.charge)
            feature.created_at = "forest"
            feature.insert(peak, scan_time)
            self.insert_feature(feature, index)
        self.count += 1

    @classmethod
    def from_reader(cls, reader, error_tolerance=1e-5, drift_error_tolerance=0.1, minimum_mass=160,
                    minimum_intensity=10., maximum_mass=float('inf'), start_time=None, end_time=None,
                    ms_level=1):

        def generate():
            if start_time is not None:
                reader.start_from_scan(rt=start_time, grouped=False)
            else:
                reader.reset()
                reader.make_iterator(grouped=False)
            i = 0
            n = len(reader)
            for i, scan in enumerate(reader):
                if end_time is not None:
                    if scan.scan_time > end_time:
                        break
                logger.info("... Processed Scan %d/%d (%0.2f%%)" % (i, n, i * 100.0 / n))
                if scan.ms_level == ms_level:
                    yield scan

        self = cls(error_tolerance=error_tolerance,
                   drift_error_tolerance=drift_error_tolerance)
        self.aggregate_peaks(
            generate(),
            minimum_mass=minimum_mass,
            minimum_intensity=minimum_intensity,
            maximum_mass=maximum_mass)
        return self



import os
import re
import time
import logging
import hashlib
import json

from collections import defaultdict
from threading import RLock, Thread, local

import idzip

import ms_deisotope

from ms_deisotope.utils import LRUDict
from ms_deisotope.data_source._compression import get_opener
from ms_deisotope.data_source.usi import USI


BLOCK_SIZE = 2 ** 20
logger = logging.getLogger("proxi_backend")


class PROXIDatasetAPIMixinBase(object):

    def _datafile_uri(self, dataset_id, data_file):
        raise NotImplementedError()

    def _is_peaklist(self, d):
        base = os.path.basename(d)
        if base.endswith(".gz"):
            base = base[:-3]
        base = base.lower()
        _, ext = base.rsplit(".", 1)
        if ext in tuple(map(str.lower, self.peaklist_file_types)):
            return True
        return False

    @property
    def peaklist_file_types(self):
        '''Alias to allow :attr:`valid_file_types` to vary.

        Returns
        -------
        list[str]'''
        return self.valid_file_types

    def describe_dataset(self, dataset_id):
        record = {
            "accession": dataset_id,
            "title": dataset_id,
            "summary": dataset_id,
            "instruments": [],
            "identifiers": [],
            "species": [],
            "modifications": [],
            "contacts": [],
            "publications": [],
            "keywords": [],
            "datasetLink": None,
            "dataFiles": [],
            "links": [],
        }
        data_files = self._find_mz_files(self.index[dataset_id])
        for d in data_files:
            if self._is_peaklist(d):
                rep = {"accession": "MS:1002850", "name": "Peak list file URI",
                       "value": self._datafile_uri(dataset_id, d)}
            else:
                rep = {"accession": "MS:1002846", "name": "Associated raw file URI",
                       "value": self._datafile_uri(dataset_id, d)}
            record['dataFiles'].append(rep)
        return record

    def enumerate_datasets(self, result_type=None):
        if result_type == 'compact':
            result = list(self.index.keys())
            return result
        return [
            self.describe_dataset(d) for d in self.index.keys()]


class IndexManagingMixin(object):

    def __init__(self, **kwargs):
        self.index = self.build_index()
        self.monitor_thread = self.monitor_index(
            kwargs.get("update_frequency"))

    def _monitor_index(self, update_frequency=None):
        if update_frequency is None:
            # Default to one hour, 60^2 seconds
            update_frequency = 60.0 ** 2
        logger.info(
            "Monitoring archive for changes every %0.2f seconds.", update_frequency)
        while True:
            time.sleep(update_frequency)
            logger.info("Rebuilding Index!")
            self.index = self.build_index(use_cache=False)

    def monitor_index(self, update_frequency=None):
        update_thread = Thread(
            target=self._monitor_index,
            args=(update_frequency, ))
        update_thread.daemon = True
        update_thread.start()
        return update_thread

    def _checksum(self, string):
        return hashlib.new("md5", string).hexdigest()

    def _find_index_cache(self, use_cache=True):
        cache_name = hashlib.md5(
            self.base_uri.encode('utf8')).hexdigest() + '.json'
        if os.path.exists(cache_name) and use_cache:
            logger.info("Reading cache from %r", cache_name)
            try:
                with open(cache_name, 'rt') as fh:
                    index = defaultdict(list, json.load(fh))
                    return cache_name, index
            except IOError:
                logger.error("Failed to read index from disk", exc_info=True)
        return cache_name, None


class SpectrumSerializerServiceMixin(object):

    def build_response(self, usi, scan):
        payload = {
            "usi": str(usi),
            "status": "READABLE",
            "attributes": [
                {"accession": "MS:1000511", "name": "ms level",
                    "value": scan.ms_level},
                {"accession": "MS:1008025", "name": "scan number",
                    "value": scan.index + 1},
                {"accession": "MS:1003061", "name": "spectrum name", "value": scan.id},
            ]
        }
        payload = self._write_peak_arrays(usi, scan, payload)
        if scan.ms_level > 1:
            mz = None
            charge = None
            if scan.precursor_information:
                mz = scan.precursor_information.mz
                charge = scan.precursor_information.charge
            elif scan.isolation_window:
                mz = scan.isolation_window.target
            if mz:
                payload['attributes'].append({
                    "accession": "MS:1000827", "name": "isolation window target m/z",
                    "value": mz,
                })
            if charge:
                payload['attributes'].append({
                    "accession": "MS:1000041", "name": "charge state",
                    "value": charge,
                })
        return payload

    def _write_peak_arrays(self, usi, scan, payload):
        mzs = []
        intensities = []
        charges = None
        base_peak = None
        nb_peaks = None
        if scan.deconvoluted_peak_set is not None:
            mzs = [p.mz for p in scan.deconvoluted_peak_set]
            intensities = [p.intensity for p in scan.deconvoluted_peak_set]
            charges = [p.charge for p in scan.deconvoluted_peak_set]
            is_profile = False
            nb_peaks = len(scan.deconvoluted_peak_set)
            base_peak = scan.base_peak.deconvoluted().intensity
        elif scan.peak_set is not None:
            mzs = [p.mz for p in scan.peak_set]
            intensities = [p.intensity for p in scan.peak_set]
            charges = None
            is_profile = False
            nb_peaks = len(scan.peak_set)
            base_peak = scan.base_peak.centroided().intensity
        else:
            mzs = scan.arrays.mz.tolist()
            intensities = scan.arrays.intensity.tolist()
            is_profile = scan.is_profile
            nb_peaks = len(mzs)
            base_peak = scan.base_peak.raw().intensity

        payload['m/z array'] = mzs
        payload['intensity array'] = intensities
        if charges is not None:
            payload['charge array'] = charges
        payload['attributes'].append(
            {"accession": "MS:1000128", "name": "profile spectrum"} if is_profile else
            {"accession": "MS:1000127", "name": "centroid spectrum"})
        payload['attributes'].append({
            "accession": 'MS:1000505', 'name': 'base peak intensity', 'value': base_peak
        })
        payload['attributes'].append({
            "accession": 'MS:1003059', 'name': 'number of peaks', 'value': nb_peaks
        })
        return payload


class DatasetNotAvailable(ValueError):
    pass


class UnrecognizedIndexFlag(ValueError):
    pass


class MassSpectraDataArchiveBase(SpectrumSerializerServiceMixin, PROXIDatasetAPIMixinBase, IndexManagingMixin):
    valid_file_types = ['mzML', 'mzXML', 'mgf', 'mzMLb']
    special_openers = set()

    def __init__(self, base_uri, cache_size=None, **kwargs):
        if cache_size is None:
            cache_size = 1
        self.base_uri = base_uri
        self.cache = LRUDict(cache_size=cache_size)
        IndexManagingMixin.__init__(self, **kwargs)

    def build_index(self, use_cache=True):
        file_types = self.valid_file_types + ['json']

        cache_name, index = self._find_index_cache(use_cache=use_cache)
        if index is not None and use_cache:
            return index

        index = defaultdict(list)

        self._traverse_files(file_types, index)

        reduced = {
            k: self._find_mz_files(v) for k, v in index.items()
        }

        n_datasets = len(reduced)
        n_ms_files = 0
        for _, v in reduced.items():
            n_ms_files += len(v)

        logger.info(
            "Collected %d Datasets with %d MS Data Files in Index", n_datasets, n_ms_files)
        logger.info("Saving index to %r", cache_name)
        try:
            with open(cache_name, 'wt') as fh:
                json.dump(index, fh, indent=2, sort_keys=True)
        except IOError:
            logger.error("Failed to write index to disk", exc_info=True)
        return index

    def _traverse_files(self, file_types, index):
        raise NotImplementedError()

    def find_ms_files_for(self, dataset, run_name):
        result = []
        members = self.index.get(dataset)
        if run_name is None:
            return members

        if members is None:
            raise DatasetNotAvailable((dataset, run_name))

        for member in members:
            base = member.split("/")[-1]
            if base.startswith(run_name):
                result.append(member)
            elif base.endswith("json"):
                if base.startswith(run_name.rsplit(".", 1)[0]):
                    result.append(member)
        # Validate that there is only one MS file for the name
        i = 0
        for res in sorted(result):
            if re.search(r"-byte-offsets.json$", res):
                continue
            if re.search(r"-idx.json$", res):
                continue
            i += 1
        if i > 1:
            raise ValueError("Multiple MS Data Files Found")
        return self._find_mz_files(result, first=True), self._find_index_files(result, first=True)

    def _find_mz_files(self, namelist, first=False):
        pattern = re.compile(r"(?:%s)(?:.gz)?$" %
                             "|".join(self.valid_file_types))
        names = []
        for name in namelist:
            if pattern.search(name):
                if first:
                    return name
                names.append(name)
        if first and not names:
            return None
        return names

    def _find_index_files(self, namelist, first=False):
        names = []
        for name in namelist:
            if re.search(r"-byte-offsets.json$", name):
                if first:
                    return name
                names.append(name)
        if first and not names:
            return None
        return names

    def _opener(self, uri, block_size=None, **kwargs):
        raise NotImplementedError()

    def open_ms_file(self, ms_file_uri, index_uri=None):
        if ms_file_uri in self.cache:
            logger.info("Found %r in the cache", ms_file_uri)
            return self.cache[ms_file_uri]
        logger.info("Opening %r", ms_file_uri)
        base_name =  ms_file_uri.split("/")[-1]
        if base_name.split('.')[-1] in self.special_openers:
            reader = ms_deisotope.MSFileLoader(ms_file_uri)
        else:
            mzml_fh = self._opener(ms_file_uri)
            if ms_file_uri.endswith("gz"):
                mzml_fh = idzip.IdzipFile(fileobj=mzml_fh)
            index_fh = None
            if index_uri is not None:
                index_fh = self._opener(index_uri, block_size=BLOCK_SIZE)
            else:
                logger.warning("Byte offset index is missing for %r", ms_file_uri)
            reader = ms_deisotope.MSFileLoader(mzml_fh, index_file=index_fh)
        logger.info("Finished opening %r", ms_file_uri)
        lock = RLock()
        self.cache[ms_file_uri] = (reader, lock)
        return reader, lock

    def list_collections(self):
        return list(self.index.keys())

    def list_datasets(self):
        return {
            k: [f.split("/")[-1] for f in self._find_mz_files(v)] for k, v in self.index.items()
        }

    def get_scan_from_usi(self, usi, reader):
        if usi.scan_identifier_type == 'scan':
            index = int(usi.scan_identifier) - 1
        elif usi.scan_identifier_type == "index":
            index = int(usi.scan_identifier)
        else:
            raise UnrecognizedIndexFlag(usi.scan_identifier_type)
        scan = reader.get_scan_by_index(index)
        return scan

    def process_scan(self, scan, reader):
        # No processing
        return scan

    def enumerate_spectra_for(self, dataset, ms_run=None):
        raise NotImplementedError()

    def handle_usi(self, usi, convert_json=True):
        '''Receive a USI, parse it into its constituent pieces, locate
        the relevant spectrum, and return it formatted message.

        Parameters
        ----------
        usi : str
            The identifier of the spectrum to be retrieved
        convert_json : bool, optional
            Whether to format the spectrum as JSON or not. Defaults to :const:`True`.

        Returns
        -------
        spectrum : dict or Scan
            The spectrum, rendered according to the `convert_json` or not.
        '''
        usi = USI.parse(str(usi))
        mzml_uri, index_uri = self.find_ms_files_for(usi.dataset, usi.datafile)
        reader, lock = self.open_ms_file(mzml_uri, index_uri)
        with lock:
            scan = self.get_scan_from_usi(usi, reader)
            scan = self.process_scan(scan, reader)
            if convert_json:
                payload = self.build_response(usi, scan)
                return payload
            else:
                return scan

    def __call__(self, usi):
        return self.handle_usi(usi)


class FSMassSpectraDataArchive(MassSpectraDataArchiveBase):
    def __init__(self, base_uri, cache_size=None, **kwargs):
        self.valid_file_types = self.valid_file_types[:]
        from ms_deisotope.data_source.thermo_raw_net import determine_if_available
        if determine_if_available():
            self.valid_file_types.append("raw")
            self.special_openers = set(self.special_openers) | {"raw"}
        super(FSMassSpectraDataArchive, self).__init__(
            base_uri, cache_size=cache_size, **kwargs)

    def _opener(self, uri, block_size=None, **kwargs):
        return open(uri, 'rb')

    def _datafile_uri(self, dataset_id, data_file):
        return os.path.normpath('/'.join([self.base_uri, data_file])).replace(os.sep, '/')

    def _traverse_files(self, file_types, index):
        for prefix, dirs, files in os.walk(self.base_uri):
            if os.sep in prefix:
                tokens = prefix.split(os.sep)
            else:
                tokens = prefix.split("/")
            if tokens[-1] == '':
                base = tokens[-2]
            else:
                base = tokens[-1]

            for f in files:
                try:
                    ext = f.rsplit(".", 1)[1]
                except IndexError:
                    continue
                if ext == 'gz':
                    ext = f.rsplit(".", 2)[1]
                if ext in file_types:
                    index[base].append(
                        os.path.normpath(
                            os.path.join(prefix, f)).replace(
                                os.sep, "/"))


class S3MassSpectraDataArchive(MassSpectraDataArchiveBase):
    def __init__(self, base_uri, cache_size=None, s3_args=None, **kwargs):
        if s3_args is None:
            s3_args = {}
        self._tls = local()
        self.s3_args = s3_args
        super(S3MassSpectraDataArchive, self).__init__(
            base_uri=base_uri, cache_size=cache_size, **kwargs)

    @property
    def s3_service(self):
        try:
            return self._tls.s3_service
        except (AttributeError, KeyError):
            from s3fs import S3FileSystem
            self._tls.s3_service = S3FileSystem(**self.s3_args)
            return self._tls.s3_service

    def _opener(self, uri, block_size=None, **kwargs):
        if block_size is None:
            block_size = BLOCK_SIZE
        return self.s3_service.open(uri, block_size=block_size)

    def _datafile_uri(self, dataset_id, data_file):
        return "s3://" + data_file

    def _traverse_files(self, file_types, index):
        for prefix, _dirs, files in self.s3_service.walk(self.base_uri):
            for f in files:
                query = f
                if query.endswith(".gz"):
                    query = query[:-3]
                ext = query.rsplit(".", 1)[-1]
                for t in file_types:
                    if t == ext:
                        index[prefix.replace(self.base_uri[5:], '')].append(
                            prefix + '/' + f)
                        break
        for group, members in list(index.items()):
            if "/" in group:
                base, _rest = group.split("/", 1)
                index[base].extend(members)
                index.pop(group)


class PeakPickingScanProcessingMixin(object):
    def __init__(self, *args, **kwargs):
        super(PeakPickingScanProcessingMixin, self).__init__(*args, **kwargs)

    def process_scan(self, scan, reader):
        scan = super(PeakPickingScanProcessingMixin,
                     self).process_scan(scan, reader)
        scan.pick_peaks()
        return scan


class DeconvolutingScanProcessingMixin(PeakPickingScanProcessingMixin):
    def __init__(self, *args, **kwargs):
        self.deconvolution_args = kwargs.get('deconvolution_args', {})
        self.deconvolution_args.setdefault("averagine", ms_deisotope.peptide)
        self.deconvolution_args.setdefault(
            "scorer", ms_deisotope.MSDeconVFitter(0))
        self.deconvolution_args.setdefault("truncate_after", 0.8)
        super(DeconvolutingScanProcessingMixin, self).__init__(*args, **kwargs)

    def process_scan(self, scan, reader):
        scan = super(DeconvolutingScanProcessingMixin, self).process_scan(scan, reader)
        scan.deconvolute(**self.deconvolution_args)
        return scan
